/*
	Genetic Algorithm Simulation for TalesWeaver
	Status Module
*/
////////////////////////////////////////////////////////////////////////

#include "../common/stdafx.h"
#include "GAstatus.h"

CGAparam *CGAstatus::m_param = NULL;

////////////////////////////////////////////////////////////////////////

// Xe[^X
void CGAstatus::init(bool do_count)
{
	m_level = D_LEVEL_START;
	for (int i=0; i<NUM_STAT; i++) m_stat[i] = D_INIT_STAT[i];
	m_hp = 0;
	m_mp = 0;
	m_point = D_INIT_POINT;
	m_cur_xien = D_SIM_INIT_XIEN;
	m_cur_stat = INDEX_S;
	m_xien_num = 0;
	m_penalty = 0;
	m_count = do_count;
	memset(m_count_xien, 0, sizeof(m_count_xien));
	memset(m_count_stat, 0, sizeof(m_count_stat));
	current_point_reset();
}

// xAbv/Xe[^XUP
bool CGAstatus::level_up()
{
	if (m_level == D_LEVEL_FINAL) return false;

	m_level++;
	m_point += m_param->lvup_point(m_level);
	m_stat[m_param->xien_up_fix(m_cur_xien)] += 100;
	m_stat[m_param->xien_up_rnd(m_cur_xien)] += D_OPT_RANDOM_RATE;
	m_hp += m_param->m_hmsw_sum_auto[INDEX_HP][m_cur_xien];
	m_mp += m_param->m_hmsw_sum_auto[INDEX_MP][m_cur_xien];
	if (m_count) m_count_xien[m_cur_xien]++;

	current_point_reset();
	return true;
}

// 蓮Xe[^XUP
bool CGAstatus::status_up()
{
	int need_point = current_point();
	if (m_point < need_point) return false;

	m_point -= need_point;
	m_stat[m_cur_stat] += 100;
	m_hp += m_param->m_hmsw_sum_manual[INDEX_HP][m_cur_stat];
	m_mp += m_param->m_hmsw_sum_manual[INDEX_MP][m_cur_stat];
	if (m_count) m_count_stat[m_cur_stat]++;

	current_point_reset();
	return true;
}

// sXienύX
void CGAstatus::change_xien(int xien)
{
	if (m_cur_xien != xien) {
		// s\ȐsXien̕ύX`FbN
		if (xien < m_cur_xien) m_xien_num++;
		if (m_xien_num < NUM_SKILL && m_level < D_SIM_SKILL[xien][m_xien_num]) m_penalty++;
	}
	m_cur_xien = xien;
}

// Xe[^X㏸ڕύX
void CGAstatus::change_stat(int stat)
{
	if (m_cur_stat != stat) current_point_reset();
	m_cur_stat = stat;
}

// Xe[^X̃I[o[t[`FbN
void CGAstatus::check_overflow()
{
	for (int i=0; i<NUM_STAT; i++) {
		if (m_stat[i] > MAX_STATUS * 100) m_stat[i] = MAX_STATUS * 100;
	}
}

// KxZo
int CGAstatus::calc_score()
{
	int score = 0;
	score += calc_score_stat();
	score += calc_score_hpmp();
	score += calc_score_point();
	score -= m_penalty * 10000;
	return score;
}

// KxZo(Xe[^X)
// Xe[^X1xdvx1œKx100Z
int CGAstatus::calc_score_stat()
{
	int score = 0;
	for (int i=0; i<NUM_STAT; i++) if (D_STAT_WEI[i]) {
		if      (m_stat[i] < D_STAT_MIN_100[i]) score += (m_stat[i]         - D_STAT_MIN_100[i]) * 10;
		else if (m_stat[i] < D_STAT_MAX_100[i]) score += (m_stat[i]         - D_STAT_MIN_100[i]) * D_STAT_WEI[i];
		else                                    score += (D_STAT_MAX_100[i] - D_STAT_MIN_100[i]) * D_STAT_WEI[i];
	}
	return score;
}

// KxZo(őHP/MP)
// HP/MP1xdvx1œKx100Z
int CGAstatus::calc_score_hpmp()
{
	int score_hp = (int)((m_hp * 50) * D_STAT_WEI_HP);
	int score_mp = (int)((m_mp * 50) * D_STAT_WEI_MP);
	return score_hp + score_mp;
}

// KxZo(c|Cg)
// c|Cg1œKx10(Xe[^X0.1xdvx1)Z
int CGAstatus::calc_score_point()
{
	return m_point * 10;
}

// HP/MP/SP/WTmzvZ({)
void CGAstatus::hmsw_distribution(int *hmsw)
{
	int hmsw_limit[4] = {20000, 5000, 9999, 9999};
	memset(hmsw, 0, sizeof(int) * NUM_HMSW * 5);

	for (int i=0; i<NUM_HMSW; i++) {
		// Up
		int v_min[NUM_XIEN];
		int v_max[NUM_XIEN];
		for (int j=0; j<NUM_XIEN; j++) {
			v_min[j] = m_param->xien_hmsw_min(j, i);
			v_max[j] = m_param->xien_hmsw_max(j, i);
			hmsw[i * 5 + INDEX_HMSW_MIN] += v_min[j] * m_count_xien[j];
			hmsw[i * 5 + INDEX_HMSW_MAX] += v_max[j] * m_count_xien[j];
		}
		int out_min, out_max;
		hmsw_distribution_random(NUM_XIEN, v_min, v_max, m_count_xien, 99.99, out_min, out_max);
		hmsw[i * 5 + INDEX_HMSW_L99] = out_min;
		hmsw[i * 5 + INDEX_HMSW_H99] = out_max;
		// 蓮Up
		for (int j=0; j<NUM_STAT; j++) {
			int v = m_param->stat_hmsw(j, i) * m_count_stat[j];
			hmsw[i * 5 + INDEX_HMSW_MIN] += v;
			hmsw[i * 5 + INDEX_HMSW_MAX] += v;
			hmsw[i * 5 + INDEX_HMSW_L99] += v;
			hmsw[i * 5 + INDEX_HMSW_H99] += v;
		}
		// l
		if (!D_INIT_USE) {
			int v = m_param->style_hmsw(D_STYLE, i);
			hmsw[i * 5 + INDEX_HMSW_MIN] += v;
			hmsw[i * 5 + INDEX_HMSW_MAX] += v;
			hmsw[i * 5 + INDEX_HMSW_L99] += v;
			hmsw[i * 5 + INDEX_HMSW_H99] += v;
		}
		// ϒl
		hmsw[i * 5 + INDEX_HMSW_AVG] = (hmsw[i * 5 + INDEX_HMSW_MIN] + hmsw[i * 5 + INDEX_HMSW_MAX]) / 2;
		// l
		if (hmsw[i * 5 + INDEX_HMSW_MIN] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_MIN] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_MAX] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_MAX] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_L99] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_L99] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_H99] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_H99] = hmsw_limit[i];
		if (hmsw[i * 5 + INDEX_HMSW_AVG] > hmsw_limit[i]) hmsw[i * 5 + INDEX_HMSW_AVG] = hmsw_limit[i];
	}
}

// HP/MP/SP/WTmzvZ(_z)
void CGAstatus::hmsw_distribution_random(int v_size, int *v_min, int *v_max, int *v_num, double dist_range, int &out_min, int &out_max)
{
	int i, j;
	int loop1, loop2;

	// vZƊme[u쐬
	int limit = 0;
	for (i=0; i<v_size; i++) limit += v_max[i] * v_num[i];
	double *dist = new double[limit + 1];
	dist[0] = 1.0;
	for (i=1; i<=limit; i++) dist[i] = 0.0;
	int dist_min = 0;
	int dist_max = 0;

	// zvZ
	for (loop1=0; loop1<v_size; loop1++) {
		int e_min = v_min[loop1];
		int e_max = v_max[loop1];
		int e_num = v_num[loop1];
		for (loop2=0; loop2<e_num; loop2++) {
			int dist_min_next = dist_min + e_min;
			int dist_max_next = dist_max + e_max;
			for (i=dist_max_next; i>=dist_min_next; i--) {
				double tmp = 0.0;
				int stt = i - e_max;
				int end = i - e_min;
				if (stt < dist_min) stt = dist_min;
				if (end > dist_max) end = dist_max;
				for (j=stt; j<=end; j++) tmp += dist[j];
				dist[i] = tmp / 4.0;
			}
			dist_min = dist_min_next;
			dist_max = dist_max_next;
		}
	}

	// ݌vmƐK
	for (i=dist_min+1; i<=dist_max; i++) dist[i] += dist[i-1];
	for (i=dist_min  ; i<=dist_max; i++) dist[i] /= dist[dist_max];

	// z؂o
	double dist_range_min = (1.0 - dist_range / 100.0) / 2.0;
	double dist_range_max = (1.0 + dist_range / 100.0) / 2.0;
	for (out_min=dist_min; out_min<=dist_max; out_min++) if (dist[out_min] > dist_range_min) break;
	for (out_max=dist_max; out_max>=dist_min; out_max--) if (dist[out_max] < dist_range_max) break;
	if (out_min > dist_min) out_max--;
	if (out_max < dist_max) out_max++;

	delete [] dist;
}

CGAstatusIO CGAstatus::class_export(CGAstatus *before)
{
	CGAstatusIO io;

	io.m_level = m_level;
	for (int i=0; i<NUM_STAT; i++) io.m_stat[i] = m_stat[i] / 100;
	io.m_point = m_point;
	io.m_cur_xien = m_cur_xien;
	io.m_cur_stat = m_cur_stat;
	io.m_need_point = current_point();
	io.m_changed_xien = !(before && before->m_cur_xien == m_cur_xien);
	io.m_changed_stat = !(before && before->m_cur_stat == m_cur_stat);

	return io;
}
