/*
	Genetic Algorithm Simulation for TalesWeaver
	Genom Unit Module
*/
////////////////////////////////////////////////////////////////////////

#include "../common/stdafx.h"
#include "GAunit.h"

CGAparam	*CGAunit::m_param = NULL;
ETYPE		CGAunit::m_backup_genom[NUM_ELEM];
CGAstatus	CGAunit::m_backup_status[NUM_ELEM];
int			CGAunit::m_backup_retcode[NUM_ELEM];

////////////////////////////////////////////////////////////////////////

// RXgN^
CGAunit::CGAunit()
{
	memset(m_genom, 0, sizeof(m_genom));
	m_evaled = false;
}

// Rs[RXgN^
CGAunit::CGAunit(CGAunit &src)
{
	*this = src;
}

// 
CGAunit &CGAunit::operator=(CGAunit &src)
{
	m_genom_num = src.m_genom_num;
	memcpy(m_genom, src.m_genom, sizeof(m_genom));
	m_status = src.m_status;
	m_score  = src.m_score;
	m_evaled = src.m_evaled;
	return *this;
}

// r
bool operator==(CGAunit &src1, CGAunit &src2)
{
	if (src1.m_genom_num != src2.m_genom_num) return false;
	ETYPE *seq1 = src1.m_genom;
	ETYPE *seq2 = src2.m_genom;
	for (int i=0; i<src1.m_genom_num; i++) if (*(seq1++) != *(seq2++)) return false;
	return true;
}

////////////////////////////////////////////////////////////////////////

// `vf(S_)
ETYPE CGAunit::element()
{
	switch (rand_range1(4)) {
	case 1: return element_stup();
	case 2: return element_xien();
	case 3: return element_stat();
	}
	return element_nop();
}

// `vf(Ȃ)
ETYPE CGAunit::element_nop()
{
	return ELEM_TYPE_NOP;
}

// `vf(Xe[^X㏸)
ETYPE CGAunit::element_stup()
{
	int idx = rand_bool() ? 0 : rand_range1(25);
	return conv_idx2elem_stup(idx);
}

// `vf(sXienύX)
ETYPE CGAunit::element_xien()
{
	while (1) {
		int idx = rand_range1(NUM_XIEN);
		if ((1 << idx) & D_AVAILABLE_XIEN) return conv_idx2elem_xien(idx);
	}
}

// `vf(Xe[^X㏸ڕύX)
ETYPE CGAunit::element_stat()
{
	while (1) {
		int idx = rand_range1(NUM_STAT);
		if ((1 << idx) & D_AVAILABLE_STAT) return conv_idx2elem_stat(idx);
	}
}

////////////////////////////////////////////////////////////////////////

// `q
void CGAunit::genom_init()
{
	for (int i=0; i<NUM_ELEM; i++) m_genom[i] = element();
	m_evaled = false;
}

// _
void CGAunit::genom_twice(CGAunit &src1, CGAunit &src2)
{
	int pos1;
	int pos2;
	do {
		pos1 = rand_range1(NUM_ELEM);
		pos2 = rand_range1(NUM_ELEM);
	} while (pos1 >= pos2);

	for (int i=0;    i<pos1;     i++) m_genom[i] = src1.m_genom[i];
	for (int i=pos1; i<pos2;     i++) m_genom[i] = src2.m_genom[i];
	for (int i=pos2; i<NUM_ELEM; i++) m_genom[i] = src1.m_genom[i];

	genom_mutate();
	m_evaled = false;
}

// l
void CGAunit::genom_unif(CGAunit &src1, CGAunit &src2)
{
	for (int i=0; i<NUM_ELEM; i++) m_genom[i] = rand_bool() ? src1.m_genom[i] : src2.m_genom[i];

	genom_mutate();
	m_evaled = false;
}

// zu]
void CGAunit::genom_swap(CGAunit &src)
{
	*this = src;

	int pos1;
	int pos2;
	do {
		if (rand_bool()) {
			pos1 = rand_range1(NUM_ELEM - 1);
			pos2 = pos1 + 1;
		} else {
			pos1 = rand_range1(NUM_ELEM);
			pos2 = rand_range1(NUM_ELEM);
		}
	} while (pos1 == pos2);
	genom_modify_swap_element(pos1, pos2);

	genom_mutate();
	m_evaled = false;
}

// gˑRψق
void CGAunit::genom_mutate()
{
	int mutate_rand = rand_b24();
	int mutate_count;
	for (mutate_count=0; mutate_count<NUM_ELEM; mutate_count++) {
		if (mutate_rand < D_GA_MUTATE_TABLE[mutate_count]) break;
	}

	if (mutate_count) {
		for (int i=0; i<mutate_count; i++) {
			int pos = rand_range1(NUM_ELEM);
			m_genom[pos] = element();
		}
		m_evaled = false;
	}
}

// `q̗vfւ(P)
void CGAunit::genom_modify_swap_element(int pos1, int pos2)
{
	ETYPE tmp     = m_genom[pos1];
	m_genom[pos1] = m_genom[pos2];
	m_genom[pos2] = tmp;
	m_evaled = false;
}

// `q̗vfւ(͈)
void CGAunit::genom_modify_swap_band(int pos11, int pos12, int pos21, int pos22)
{
	if (!(pos11 < pos12 && pos12 < pos21 && pos21 < pos22)) return;
	
	ETYPE tmp[NUM_ELEM];
	int pos = pos11;
	for (int i=pos21;   i<=pos22; i++) tmp[pos++] = m_genom[i];
	for (int i=pos12+1; i< pos21; i++) tmp[pos++] = m_genom[i];
	for (int i=pos11;   i<=pos12; i++) tmp[pos++] = m_genom[i];
	for (int i=pos11;   i<=pos22; i++) m_genom[i] = tmp[i];
	m_evaled = false;
}

////////////////////////////////////////////////////////////////////////

// Xe[^X(CGAstatusIO)`q(CGAunit)֕ϊ
void CGAunit::class_import(int &io_size, CGAstatusIO *io)
{
	m_evaled = false;
}

// `q(CGAunit)Xe[^X(CGAstatusIO)֕ϊ
void CGAunit::class_export(int &io_size, CGAstatusIO *io)
{
	io_size = 0;

	int i_last = -1;
	for (int i=0; i<m_genom_num; i++) {
		if (conv_elem2type(m_genom[i]) != ELEM_TYPE_STUP) continue;

		io[io_size] = m_backup_status[i].class_export(i_last >= 0 ? &m_backup_status[i_last] : NULL);
		io_size++;
		i_last = i;
	}
}

////////////////////////////////////////////////////////////////////////

#define EVAL_RETCODE_CONT 0
#define EVAL_RETCODE_CNOP 1
#define EVAL_RETCODE_QUIT 2
#define EVAL_RETCODE_QNOP 3

// Kx]
int CGAunit::eval(bool do_force, int *hmsw)
{
	int retcode;
	int retcode_list[NUM_ELEM];

	// ɕ]Ă炻̒lԂ
	if (m_evaled && !do_force) return m_score;

	// Xe[^X
	m_genom_num = 0;
	m_status.init(hmsw ? true : false);
	retcode = EVAL_RETCODE_CONT;

	// obNAbv烍[h
	if (!do_force) retcode = eval_backup_load(retcode_list);

	// xI[ɒB܂ŏԂZo
	while (retcode == EVAL_RETCODE_CONT || retcode == EVAL_RETCODE_CNOP) {
		// `vfI[ɒBĂ΋I
		if (m_genom_num == NUM_ELEM) break;
		// `vfǂݍ
		retcode = eval_accept_elem(m_genom[m_genom_num]);
		// obNAbvL^
		retcode_list[m_genom_num] = retcode;
		if (do_force) eval_backup_save(retcode);
		// ̈`vf
		m_genom_num++;
	}
	if (retcode == EVAL_RETCODE_QNOP) {
		while (retcode_list[((--m_genom_num)-1)] != EVAL_RETCODE_CONT);
	}

	// I[o[t[
	m_status.check_overflow();

	// HP/MP/SP/WTmzvZ
	if (hmsw) m_status.hmsw_distribution(hmsw);

	// ]
	m_score = m_status.calc_score() + eval_calc_score();
	m_evaled = true;
	return m_score;
}

// Xe[^XZ[u
void CGAunit::eval_backup_save(int retcode)
{
	m_backup_genom[m_genom_num] = m_genom[m_genom_num];
	m_backup_status[m_genom_num] = m_status;
	m_backup_retcode[m_genom_num] = retcode;
}

// `qʂĂ͈͂ł̃Xe[^X[h
int CGAunit::eval_backup_load(int *retcode_list)
{
	for (m_genom_num=0; m_genom_num<NUM_ELEM; m_genom_num++) {
		if (m_genom[m_genom_num] != m_backup_genom[m_genom_num]) break;
		if (m_backup_retcode[m_genom_num] == EVAL_RETCODE_QUIT || m_backup_retcode[m_genom_num] == EVAL_RETCODE_QNOP) {
			m_genom_num++;
			break;
		}
	}
	if (m_genom_num == 0) return EVAL_RETCODE_CONT;
	m_status = m_backup_status[m_genom_num-1];
	for (int i=0; i<m_genom_num; i++) retcode_list[i] = m_backup_retcode[i];
	return retcode_list[m_genom_num-1];
}

// `q1Kp
int CGAunit::eval_accept_elem(ETYPE elem)
{
	ETYPE elem_type = conv_elem2type(elem);
	int elem_idx = conv_elem2idx(elem);

	switch (elem_type) {
		case ELEM_TYPE_NOP:
			// Ȃ
			return EVAL_RETCODE_CNOP;
		case ELEM_TYPE_STUP:
			// Xe[^XKl㏸邩Kv|Cgオ܂ŋɐU
			return eval_up_to_next(elem_idx);
		case ELEM_TYPE_XIEN:
			// sXienύX
			m_status.change_xien(elem_idx);
			return EVAL_RETCODE_CNOP;
		case ELEM_TYPE_STAT:
			// Xe[^X㏸ڕύX
			m_status.change_stat(elem_idx);
			return EVAL_RETCODE_CNOP;
	}
	return EVAL_RETCODE_CONT;
}

// Xe[^XKl㏸邩Kv|Cgオ܂ŋɐU
int CGAunit::eval_up_to_next(int next)
{
	int next_status = m_status.current_status() + next * 100;
	int next_point  = m_status.current_point() + 1;

	bool up = false;
	while (1) {
		// 蓮Xe[^XUP
		bool do_manual = m_status.status_up();
		// xAbv/Xe[^XUP
		if (!do_manual) {
			bool do_auto = m_status.level_up();
			if (!do_auto) return up ? EVAL_RETCODE_QUIT : EVAL_RETCODE_QNOP;
		}
		up = true;
		// I
		if (m_status.current_point()  >= next_point) break;
		if (m_status.current_status() >= next_status && next) break;
	}
	return EVAL_RETCODE_CONT;
}

// `qR̃XRA
int CGAunit::eval_calc_score()
{
	int score = 0;

	for (int i=0; i<m_genom_num; i++) {
		ETYPE elem_type = conv_elem2type(m_genom[i]);
		int elem_idx = conv_elem2idx(m_genom[i]);
		switch (elem_type) {
		case ELEM_TYPE_STUP:
			if (elem_idx) score--;
			break;
		case ELEM_TYPE_XIEN:
			score -= D_OPT_PENALTY_XIEN;
			break;
		case ELEM_TYPE_STAT:
			score -= D_OPT_PENALTY_STAT;
			break;
		}
	}

	return score;
}
